@@ -18,7 +18,7 @@ from django.utils.html import conditional_escape, format_html
 from django.utils.safestring import mark_safe
 from django.db.models import QuerySet
 import six
-from tri.declarative import declarative, creation_ordered, with_meta, setdefaults, evaluate_recursive, evaluate, getattr_path, sort_after, LAST, setdefaults_path, dispatch, EMPTY, Namespace, setattr_path, RefinableObject, refinable, Refinable, shortcut, Shortcut
+from tri.declarative import declarative, creation_ordered, with_meta, evaluate_recursive, evaluate, getattr_path, sort_after, LAST, setdefaults_path, dispatch, EMPTY, Namespace, setattr_path, RefinableObject, refinable, Refinable, shortcut, Shortcut
 from tri.form import Field, Form, member_from_model, expand_member, create_members_from_model, render_template, handle_dispatch, DISPATCH_PATH_SEPARATOR, evaluate_and_group_links
 from tri.named_struct import NamedStructField, NamedStruct
 from tri.struct import Struct, merged
@@ -27,7 +27,7 @@ from tri.form import Link as tri_form_Link
 
 from tri.table.db_compat import setup_db_compat
 
-__version__ = '5.3.0'  # pragma: no mutate
+__version__ = '6.0.1'  # pragma: no mutate
 
 LAST = LAST
 
@@ -38,25 +38,33 @@ def register_column_factory(field_class, factory):
     _column_factory_by_field_type[field_class] = factory
 
 
-def prepare_headers(request, bound_columns):
+def _with_path_prefix(table, name):
+    if table.name:
+        return DISPATCH_PATH_SEPARATOR.join([table.name, name])
+    else:
+        return name
+
+
+def prepare_headers(table, bound_columns):
     """
     :type bound_columns: list of BoundColumn
     """
-    if request is None:
+    if table.request is None:
         return
 
     for column in bound_columns:
         if column.sortable:
-            params = request.GET.copy()
-            order = request.GET.get('order', None)
+            params = table.request.GET.copy()
+            param_path = _with_path_prefix(table, 'order')
+            order = table.request.GET.get(param_path, None)
             start_sort_desc = column.sort_default_desc
-            params['order'] = column.name if not start_sort_desc else '-' + column.name
+            params[param_path] = column.name if not start_sort_desc else '-' + column.name
             if order is not None:
                 is_desc = len(order) > 0 and order[0] == '-'
                 order_field = is_desc and order[1:] or order
                 if order_field == column.name:
                     new_order = is_desc and order[1:] or "-%s" % order
-                    params['order'] = new_order
+                    params[param_path] = new_order
             column.is_sorting = False if order is None else (column.name == order or ('-' + column.name) == order)
             column.url = "?%s" % params.urlencode()
         else:
@@ -311,7 +319,7 @@ def column_shortcut_icon(icon, is_report=False, icon_title='', show=True, call_t
 
     :param icon: the font awesome name of the icon
     """
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         show=lambda table, **rest: evaluate(show, table=table, **rest) and not is_report,
         title=icon_title,
         cell__format=lambda value, **_: mark_safe('<i class="fa fa-lg fa-%s"%s></i>' % (icon, ' title="%s"' % icon_title if icon_title else '')) if value else ''
@@ -384,7 +392,7 @@ def column_shortcut_run(is_report=False, show=True, call_target=None, **kwargs):
     """
     Shortcut for creating a clickable run icon. The URL defaults to `your_object.get_absolute_url() + 'run/'`. Specify the option cell__url to override.
     """
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         show=lambda table, **rest: evaluate(show, table=table, **rest) and not is_report,
     ))
     return call_target(**kwargs)
@@ -393,12 +401,15 @@ def column_shortcut_run(is_report=False, show=True, call_target=None, **kwargs):
 Column.run = staticmethod(column_shortcut_run)
 
 
+SELECT_DISPLAY_NAME = '<i class="fa fa-check-square-o" onclick="tri_table_js_select_all(this)"></i>'
+
+
 @shortcut
 @dispatch(
     call_target=Column,
     name='__select__',
     title='Select all',
-    display_name=mark_safe('<i class="fa fa-check-square-o"></i>'),
+    display_name=mark_safe(SELECT_DISPLAY_NAME),
     sortable=False,
     attrs__class__thin=True,
     attrs__class__nopad=True,
@@ -411,7 +422,7 @@ def column_shortcut_select(is_report=False, checkbox_name='pk', show=True, check
     :param checkbox_name: the name of the checkbox. Default is "pk", resulting in checkboxes like "pk_1234".
     :param checked: callable to specify if the checkbox should be checked initially. Defaults to False.
     """
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         show=lambda table, **rest: evaluate(show, table=table, **rest) and not is_report,
         cell__value=lambda row, **_: mark_safe('<input type="checkbox"%s class="checkbox" name="%s_%s" />' % (' checked' if checked(row.pk) else '', checkbox_name, row.pk)),
     ))
@@ -437,13 +448,17 @@ def column_shortcut_boolean(is_report=False, call_target=None, **kwargs):
             value = value()
         return mark_safe('<i class="fa fa-check" title="Yes"></i>') if value else ''
 
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         cell__format=lambda value, **rest: yes_no_formatter(value=value, **rest) if is_report else render_icon(value),
     ))
     return call_target(**kwargs)
 
 
 Column.boolean = staticmethod(column_shortcut_boolean)
+Column.boolean_tristate = Shortcut(
+    query__call_target=Variable.boolean_tristate,
+    call_target=Column.boolean,
+)
 
 
 @shortcut
@@ -454,7 +469,7 @@ Column.boolean = staticmethod(column_shortcut_boolean)
 )
 def column_shortcut_choice(call_target, **kwargs):
     choices = kwargs['choices']
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         bulk__choices=choices,
         query__choices=choices,
     ))
@@ -471,7 +486,7 @@ Column.choice = staticmethod(column_shortcut_choice)
     query__call_target=Variable.choice_queryset,
 )
 def column_shortcut_choice_queryset(call_target, **kwargs):
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         bulk__model=kwargs.get('model'),
         query__model=kwargs.get('model'),
     ))
@@ -489,7 +504,7 @@ Column.choice_queryset = staticmethod(column_shortcut_choice_queryset)
     cell__format=lambda value, **_: ', '.join(['%s' % x for x in value.all()]),
 )
 def column_shortcut_multi_choice_queryset(call_target, **kwargs):
-    setdefaults(kwargs, dict(
+    setdefaults_path(kwargs, dict(
         bulk__model=kwargs.get('model'),
         query__model=kwargs.get('model'),
     ))
@@ -593,7 +608,7 @@ class BoundRow(object):
 
     def render(self):
         if self.template:
-            context = dict(bound_row=self, row=self.row, table=self.table)
+            context = dict(bound_row=self, row=self.row, **self.table.context)
             return render_template(self.table.request, self.template, context)
 
         return format_html('<tr{}>{}</tr>', self.render_attrs(), self.render_cells())
@@ -708,15 +723,20 @@ class Table(RefinableObject):
     .. code:: python
 
         class FooTable(Table):
-            class Meta:
-                sortable = False
             a = Column()
             b = Column()
 
+            class Meta:
+                sortable = False
+                attrs__style = 'background: green'
+
     """
 
+    name = Refinable()
     bulk_filter = Refinable()
+    """ :type: tri.declarative.Namespace """
     bulk_exclude = Refinable()
+    """ :type: tri.declarative.Namespace """
     sortable = Refinable()
     default_sort_order = Refinable()
     attrs = Refinable()
@@ -732,10 +752,16 @@ class Table(RefinableObject):
     """ :type: tri.declarative.Namespace """
     endpoint = Refinable()
 
+    @staticmethod
+    @refinable
+    def preprocess_data(data, **_):
+        return data
+
     @staticmethod
     @refinable
     def preprocess_row(table, row, **_):
-        return None
+        del table
+        return row
 
     @dispatch(
         column=EMPTY,
@@ -897,8 +923,7 @@ class Table(RefinableObject):
         if self.request is None:
             return
 
-        order = self.request.GET.get('order', self.default_sort_order)
-
+        order = self.request.GET.get(_with_path_prefix(self, 'order'), self.default_sort_order)
         if order is not None:
             is_desc = order[0] == '-'
             order_field = is_desc and order[1:] or order
@@ -922,7 +947,7 @@ class Table(RefinableObject):
                     self.data = self.data.order_by(*order_args)
 
     def _prepare_headers(self):
-        prepare_headers(self.request, self.shown_bound_columns)
+        prepare_headers(self, self.shown_bound_columns)
         # The id(header) and the type(x.display_name) stuff is to make None not be equal to None in the grouping
         group_columns = []
 
@@ -1038,6 +1063,7 @@ class Table(RefinableObject):
             variables = list(generate_variables())
 
             self._query = Query(
+                gui__name=self.name,
                 request=self.request,
                 variables=variables,
                 endpoint_dispatch_prefix=DISPATCH_PATH_SEPARATOR.join(part for part in [self.endpoint_dispatch_prefix, 'query'] if part is not None),
@@ -1048,7 +1074,9 @@ class Table(RefinableObject):
             self._query_error = ''
             if self._query_form:
                 try:
-                    self.data = self.data.filter(self.query.to_q())
+                    q = self.query.to_q()
+                    if q:
+                        self.data = self.data.filter(q)
                 except QueryException as e:
                     self._query_error = str(e)
 
@@ -1069,13 +1097,18 @@ class Table(RefinableObject):
                             bulk_namespace['field_name'] = column.attr
                         yield bulk_namespace()
             bulk_fields = list(generate_bulk_fields())
+            if bulk_fields:
+                bulk_fields.append(Field.hidden(name='_all_pks_', attr=None, initial='0', required=False, template='tri_form/input.html'))
 
-            self._bulk_form = Form(
-                data=self.request.POST,
-                fields=bulk_fields,
-                endpoint_dispatch_prefix=DISPATCH_PATH_SEPARATOR.join(part for part in [self.endpoint_dispatch_prefix, 'bulk'] if part is not None),
-                **self.bulk
-            ) if bulk_fields else None
+                self._bulk_form = Form(
+                    data=self.request.POST,
+                    fields=bulk_fields,
+                    name=self.name,
+                    endpoint_dispatch_prefix=DISPATCH_PATH_SEPARATOR.join(part for part in [self.endpoint_dispatch_prefix, 'bulk'] if part is not None),
+                    **self.bulk
+                )
+            else:
+                self._bulk_form = None
 
         self._prepare_auto_rowspan()
 
@@ -1084,8 +1117,13 @@ class Table(RefinableObject):
 
     def __iter__(self):
         self.prepare()
-        for i, row in enumerate(self.data):
-            self.preprocess_row(table=self, row=row)
+        for i, row in enumerate(self.preprocess_data(self.data)):
+            new_row = self.preprocess_row(table=self, row=row)
+            if new_row is None:
+                warnings.warn('preprocess_row must return the object that has been processed', DeprecationWarning)
+                new_row = row
+            row = new_row
+
             yield BoundRow(table=self, row=row, row_index=i, **evaluate_recursive(self.row, table=self, row=row))
 
     def render_attrs(self):
@@ -1095,6 +1133,35 @@ class Table(RefinableObject):
     def render_tbody(self):
         return mark_safe('\n'.join([bound_row.render() for bound_row in self.bound_rows()]))
 
+    def render_paginator(self, adjacent_pages=6):
+        context = self.context.copy()
+        page = context["page"]
+        assert page != 0  # pages are 1-indexed!
+        if page <= adjacent_pages:
+            page = adjacent_pages + 1
+        elif page > context["pages"] - adjacent_pages:
+            page = context["pages"] - adjacent_pages
+        page_numbers = [
+            n for n in
+            range(page - adjacent_pages, page + adjacent_pages + 1)
+            if 0 < n <= context["pages"]
+        ]
+
+        get = context['request'].GET.copy() if 'request' in context else {}
+        if 'page' in get:
+            del get['page']
+
+        context = merged(context, dict(
+            extra=get and (get.urlencode() + "&") or "",
+            page_numbers=page_numbers,
+            show_first=1 not in page_numbers,
+            show_last=context["pages"] not in page_numbers,
+            show_hits=context["show_hits"],
+            hit_label=context["hit_label"],
+        ))
+
+        return render_template(request=self.request, template='tri_table/paginator.html', context=context)
+
     @staticmethod
     @dispatch(
         column=EMPTY,
@@ -1140,6 +1207,17 @@ class Table(RefinableObject):
             if prefix == endpoint:
                 return handler(table=self, key=remaining_key, value=value)
 
+    def bulk_queryset(self):
+        queryset = self.model.objects.all() \
+            .filter(**self.bulk_filter) \
+            .exclude(**self.bulk_exclude)
+
+        if self._bulk_form.fields_by_name._all_pks_.value == '1':
+            return queryset
+        else:
+            pks = [key[len('pk_'):] for key in self.request.POST if key.startswith('pk_')]
+            return queryset.filter(pk__in=pks)
+
 
 class Link(tri_form_Link):
     """
@@ -1149,7 +1227,7 @@ class Link(tri_form_Link):
     # backwards compatibility with old interface
     def __init__(self, title, url=None, **kwargs):
         if url:
-            warnings.warn('url parameter is deprecated, use attrs__href')
+            warnings.warn('url parameter is deprecated, use attrs__href', DeprecationWarning)
             kwargs['attrs__href'] = url
 
         super(Link, self).__init__(title=title, **kwargs)
@@ -1248,7 +1326,7 @@ def render_table(request,
                  paginator=None,
                  show_hits=False,
                  hit_label='Items',
-                 post_bulk_edit=lambda table, pks, queryset, updates: None):
+                 post_bulk_edit=lambda table, queryset, updates: None):
     """
     Render a table. This automatically handles pagination, sorting, filtering and bulk operations.
 
@@ -1280,21 +1358,17 @@ def render_table(request,
     context['tri_query_error'] = table.query_error
 
     if table.bulk_form and request.method == 'POST':
-        pks = [key[len('pk_'):] for key in request.POST if key.startswith('pk_')]
-
         if table.bulk_form.is_valid():
+            queryset = table.bulk_queryset()
+
             updates = {
                 field.name: field.value
                 for field in table.bulk_form.fields
-                if field.value is not None and field.value is not ''
+                if field.value is not None and field.value is not '' and field.attr is not None
             }
-            queryset = table.model.objects.all() \
-                .filter(pk__in=pks) \
-                .filter(**table.bulk_filter) \
-                .exclude(**table.bulk_exclude)
             queryset.update(**updates)
 
-            post_bulk_edit(table=table, pks=pks, queryset=queryset, updates=updates)
+            post_bulk_edit(table=table, queryset=queryset, updates=updates)
 
             return HttpResponseRedirect(request.META['HTTP_REFERER'])
 
